/// <reference types="node" />
import { UrlReader, ContainerRunner, PluginEndpointDiscovery } from '@backstage/backend-common';
import { Config } from '@backstage/config';
import { Logger } from 'winston';
import { ScmIntegrationRegistry } from '@backstage/integration';
import { Entity, EntityName } from '@backstage/catalog-model';
import { Writable } from 'stream';
import express from 'express';
import { IndexableDocument } from '@backstage/search-common';

declare type PreparerResponse = {
    /**
     * The path to directory where the tree is downloaded.
     */
    preparedDir: string;
    /**
     * A unique identifier of the tree blob, usually the commit SHA or etag from the target.
     */
    etag: string;
};
declare type PreparerBase = {
    /**
     * Given an Entity definition from the Software Catalog, go and prepare a directory
     * with contents from the location in temporary storage and return the path.
     *
     * @param entity - The entity from the Software Catalog
     * @param options - If etag is provided, it will be used to check if the target has
     *        updated since the last build.
     * @throws `NotModifiedError` when the prepared directory has not been changed since the last build.
     */
    prepare(entity: Entity, options?: {
        logger?: Logger;
        etag?: string;
    }): Promise<PreparerResponse>;
};
declare type PreparerBuilder = {
    register(protocol: RemoteProtocol, preparer: PreparerBase): void;
    get(entity: Entity): PreparerBase;
};
declare type RemoteProtocol = 'url' | 'dir';

declare type ParsedLocationAnnotation = {
    type: RemoteProtocol;
    target: string;
};
declare const parseReferenceAnnotation: (annotationName: string, entity: Entity) => ParsedLocationAnnotation;
/**
 * TechDocs references of type `dir` are relative the source location of the entity.
 * This function transforms relative references to absolute ones, based on the
 * location the entity was ingested from. If the entity was registered by a `url`
 * location, it returns a `url` location with a resolved target that points to the
 * targeted subfolder. If the entity was registered by a `file` location, it returns
 * an absolute `dir` location.
 *
 * @param entity - the entity with annotations
 * @param dirAnnotation - the parsed techdocs-ref annotation of type 'dir'
 * @param scmIntegrations - access to the scmIntegration to do url transformations
 * @throws if the entity doesn't specify a `dir` location or is ingested from an unsupported location.
 * @returns the transformed location with an absolute target.
 */
declare const transformDirLocation: (entity: Entity, dirAnnotation: ParsedLocationAnnotation, scmIntegrations: ScmIntegrationRegistry) => {
    type: 'dir' | 'url';
    target: string;
};
declare const getLocationForEntity: (entity: Entity, scmIntegration: ScmIntegrationRegistry) => ParsedLocationAnnotation;
declare const getDocFilesFromRepository: (reader: UrlReader, entity: Entity, opts?: {
    etag?: string | undefined;
    logger?: Logger | undefined;
} | undefined) => Promise<PreparerResponse>;

/**
 * The values that the generator will receive.
 *
 * @public
 * @param inputDir - The directory of the uncompiled documentation, with the values from the frontend
 * @param outputDir - Directory to store generated docs in. Usually - a newly created temporary directory.
 * @param parsedLocationAnnotation - backstage.io/techdocs-ref annotation of an entity
 * @param etag - A unique identifier for the prepared tree e.g. commit SHA. If provided it will be stored in techdocs_metadata.json.
 * @param logger - A logger that forwards the messages to the caller to be displayed outside of the backend.
 * @param logStream - A log stream that can send raw log messages to the caller to be displayed outside of the backend.
 */
declare type GeneratorRunOptions = {
    inputDir: string;
    outputDir: string;
    parsedLocationAnnotation?: ParsedLocationAnnotation;
    etag?: string;
    logger: Logger;
    logStream?: Writable;
};
declare type GeneratorBase = {
    run(opts: GeneratorRunOptions): Promise<void>;
};
/**
 * List of supported generator options
 */
declare type SupportedGeneratorKey = 'techdocs' | string;
/**
 * The generator builder holds the generator ready for run time
 */
declare type GeneratorBuilder = {
    register(protocol: SupportedGeneratorKey, generator: GeneratorBase): void;
    get(entity: Entity): GeneratorBase;
};

declare class TechdocsGenerator implements GeneratorBase {
    /**
     * The default docker image (and version) used to generate content. Public
     * and static so that techdocs-common consumers can use the same version.
     */
    static readonly defaultDockerImage = "spotify/techdocs:v0.3.6";
    private readonly logger;
    private readonly containerRunner;
    private readonly options;
    private readonly scmIntegrations;
    static fromConfig(config: Config, options: {
        containerRunner: ContainerRunner;
        logger: Logger;
    }): TechdocsGenerator;
    constructor(options: {
        logger: Logger;
        containerRunner: ContainerRunner;
        config: Config;
        scmIntegrations: ScmIntegrationRegistry;
    });
    run(options: GeneratorRunOptions): Promise<void>;
}

declare class Generators implements GeneratorBuilder {
    private generatorMap;
    static fromConfig(config: Config, options: {
        logger: Logger;
        containerRunner: ContainerRunner;
    }): Promise<GeneratorBuilder>;
    register(generatorKey: SupportedGeneratorKey, generator: GeneratorBase): void;
    get(entity: Entity): GeneratorBase;
}

declare class DirectoryPreparer implements PreparerBase {
    private readonly scmIntegrations;
    private readonly reader;
    constructor(config: Config, _logger: Logger, reader: UrlReader);
    prepare(entity: Entity, options?: {
        logger?: Logger;
        etag?: string;
    }): Promise<PreparerResponse>;
}

declare class UrlPreparer implements PreparerBase {
    private readonly logger;
    private readonly reader;
    constructor(reader: UrlReader, logger: Logger);
    prepare(entity: Entity, options?: {
        etag?: string;
    }): Promise<PreparerResponse>;
}

declare type factoryOptions$1 = {
    logger: Logger;
    reader: UrlReader;
};
declare class Preparers implements PreparerBuilder {
    private preparerMap;
    static fromConfig(config: Config, { logger, reader }: factoryOptions$1): Promise<PreparerBuilder>;
    register(protocol: RemoteProtocol, preparer: PreparerBase): void;
    get(entity: Entity): PreparerBase;
}

/**
 * Key for all the different types of TechDocs publishers that are supported.
 */
declare type PublisherType = 'local' | 'googleGcs' | 'awsS3' | 'azureBlobStorage' | 'openStackSwift';
declare type PublishRequest = {
    entity: Entity;
    directory: string;
};
/**
 * Response containing metadata about where files were published and what may
 * have been published or updated.
 */
declare type PublishResponse = {
    /**
     * The URL which serves files from the local publisher's static directory.
     */
    remoteUrl?: string;
    /**
     * The list of objects (specifically their paths) that were published.
     * Objects do not have a preceding slash, and match how one would load the
     * object over the `/static/docs/*` TechDocs Backend Plugin endpoint.
     */
    objects?: string[];
} | void;
/**
 * Result for the validation check.
 *
 * @public
 */
declare type ReadinessResponse = {
    /** If true, the publisher is able to interact with the backing storage. */
    isAvailable: boolean;
};
/**
 * Type to hold metadata found in techdocs_metadata.json and associated with each site
 * @param etag - ETag of the resource used to generate the site. Usually the latest commit sha of the source repository.
 */
declare type TechDocsMetadata = {
    site_name: string;
    site_description: string;
    etag: string;
    build_timestamp: number;
    files?: string[];
};
declare type MigrateRequest = {
    /**
     * Whether or not to remove the source file. Defaults to false (acting like a
     * copy instead of a move).
     */
    removeOriginal?: boolean;
    /**
     * Maximum number of files/objects to migrate at once. Defaults to 25.
     */
    concurrency?: number;
};
/**
 * Base class for a TechDocs publisher (e.g. Local, Google GCS Bucket, AWS S3, etc.)
 * The publisher handles publishing of the generated static files after the prepare and generate steps of TechDocs.
 * It also provides APIs to communicate with the storage service.
 *
 * @public
 */
interface PublisherBase {
    /**
     * Check if the publisher is ready. This check tries to perform certain checks to see if the
     * publisher is configured correctly and can be used to publish or read documentations.
     * The different implementations might e.g. use the provided service credentials to access the
     * target or check if a folder/bucket is available.
     */
    getReadiness(): Promise<ReadinessResponse>;
    /**
     * Store the generated static files onto a storage service (either local filesystem or external service).
     *
     * @param request - Object containing the entity from the service
     *                  catalog, and the directory that contains the generated static files from TechDocs.
     */
    publish(request: PublishRequest): Promise<PublishResponse>;
    /**
     * Retrieve TechDocs Metadata about a site e.g. name, contributors, last updated, etc.
     * This API uses the techdocs_metadata.json file that co-exists along with the generated docs.
     */
    fetchTechDocsMetadata(entityName: EntityName): Promise<TechDocsMetadata>;
    /**
     * Route middleware to serve static documentation files for an entity.
     */
    docsRouter(): express.Handler;
    /**
     * Check if the index.html is present for the Entity at the Storage location.
     */
    hasDocsBeenGenerated(entityName: Entity): Promise<boolean>;
    /**
     * Migrates documentation objects with case sensitive entity triplets to
     * lowercase entity triplets. This was (will be) a change introduced in
     * `techdocs-cli` version `{0.x.y}` and `techdocs-backend` version `{0.x.y}`.
     *
     * Implementation of this method is unnecessary in publishers introduced
     * after version `{0.x.y}` of `techdocs-common`.
     */
    migrateDocsCase?(migrateRequest: MigrateRequest): Promise<void>;
}

declare type factoryOptions = {
    logger: Logger;
    discovery: PluginEndpointDiscovery;
};
/**
 * Factory class to create a TechDocs publisher based on defined publisher type in app config.
 * Uses `techdocs.publisher.type`.
 */
declare class Publisher {
    static fromConfig(config: Config, { logger, discovery }: factoryOptions): Promise<PublisherBase>;
}

interface TechDocsDocument extends IndexableDocument {
    kind: string;
    namespace: string;
    name: string;
    lifecycle: string;
    owner: string;
    path: string;
}

export { DirectoryPreparer, GeneratorBase, GeneratorBuilder, GeneratorRunOptions, Generators, ParsedLocationAnnotation, PreparerBase, PreparerBuilder, Preparers, Publisher, PublisherBase, PublisherType, ReadinessResponse, RemoteProtocol, TechDocsDocument, TechDocsMetadata, TechdocsGenerator, UrlPreparer, getDocFilesFromRepository, getLocationForEntity, parseReferenceAnnotation, transformDirLocation };
